package org.ylhealth.ym.springtest.hibernate;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import javax.persistence.EntityManager;
import javax.persistence.Id;
import javax.persistence.Query;
import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;
import org.hibernate.SQLQuery;
import org.hibernate.transform.Transformers;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.i18n.LocaleContextHolder;
import com.google.common.cache.Cache;
import com.google.common.cache.CacheBuilder;

/** Spring data 取出後進行多國語系的替換 快取多語資料 30 分鐘 */
public class I18NPostProcessor implements MethodInterceptor {
  private Logger logger = LoggerFactory.getLogger(getClass());
  private EntityManager em;

  public I18NPostProcessor(EntityManager em) {
    this.em = em;
  }

  @SuppressWarnings("rawtypes")
  private Cache<String, Map> cache =
      CacheBuilder.newBuilder().expireAfterAccess(30, TimeUnit.MINUTES).build();

  @Override
  public Object invoke(MethodInvocation invocation) throws Throwable {
    Object result = invocation.proceed();
    process(result);
    return result;
  }

  /**
   * 處理查詢結果，如果是 Collection 的話，第一筆沒處理就不處理了
   * @param result
   * @throws IllegalAccessException
   * @throws InvocationTargetException
   */
  private void process(Object result) throws IllegalAccessException, InvocationTargetException {
    if (result instanceof Collection) {
      for (Object entity : (Collection) result) {
        if (!processEntity(entity)) break;
      }
    } else {
      processEntity(result);
    }
  }

  /**
   * @param entity
   * @return　是否有處理
   * @throws IllegalAccessException
   * @throws InvocationTargetException
   */
  private boolean processEntity(Object entity)
      throws IllegalAccessException, InvocationTargetException {
    boolean result = false;
    I18NTranslates i18n = entity.getClass().getAnnotation(I18NTranslates.class);
    if (i18n != null) {
      logger.trace("I18n process: {}", entity);
      Class clazz = entity.getClass();
      Map<String, String> translate = getTranslate(i18n, entity, clazz);
      if (translate != null && !translate.isEmpty()) {
        Field[] fields = clazz.getDeclaredFields();
        for (I18NTranslate i18nTranslate : i18n.mapping()) {
          for (Field field : fields) {
            if (field.getName().equals(i18nTranslate.field())) {
              field.setAccessible(true);
              field.set(entity, translate.get(i18nTranslate.column()));
              break;
            }
          }
        }
      }
      result = true;
    }
    return result;
  }

  /**
   * @param i18n
   * @param entity
   * @param clazz
   * @return
   * @throws IllegalAccessException
   * @throws InvocationTargetException
   */
  private Map<String, String> getTranslate(I18NTranslates i18n, Object entity, Class clazz)
      throws IllegalAccessException, InvocationTargetException {
    Object id = getPKValue(clazz, entity);
    String language = LocaleContextHolder.getLocale().toString();
    String table = i18n.table();
    String idColumn = i18n.id();

    Map<String, Map<String, String>> translateTable =
        translateTable(table, language, i18n.mapping(), idColumn);

    return translateTable.get(id);
  }

  private Map<String, Map<String, String>> translateTable(
      String table, String language, I18NTranslate[] mapping, String idColumn) {
    String key = "Table:" + table + "_" + language;
    Map<String, Map<String, String>> result = cache.getIfPresent(key);
    if (result == null) {
      logger.trace("translate query, table: {}", table);
      String sql = String.format("select * from %s where languageCode = ?", table);
      Query query = em.createNativeQuery(sql);
      SQLQuery query1 = query.unwrap(org.hibernate.SQLQuery.class);
      query1.setResultTransformer(Transformers.ALIAS_TO_ENTITY_MAP);
      query1.addScalar(idColumn);
      for (I18NTranslate i18nTranslate : mapping) {
        query1.addScalar(i18nTranslate.column());
      }
      query.setParameter(1, language);
      List<Map<String, String>> result1 = query.getResultList();
      result = new HashMap<>();
      for (Map<String, String> map : result1) {
        result.put(map.get(idColumn), map);
      }
      cache.put(key, result);
    }
    return result;
  }

  private Object getPKValue(Class<?> clazz, Object entity)
      throws IllegalArgumentException, IllegalAccessException, InvocationTargetException {
    Collection<Field> fields = getAllDeclaredFields(clazz);
    for (Field field : fields) {
      if (field.getAnnotation(Id.class) != null) {
        field.setAccessible(true);
        return field.get(entity);
      }
    }
    for (Method method : clazz.getDeclaredMethods()) {
      if (method.getAnnotation(Id.class) != null) {
        return method.invoke(entity, null);
      }
    }
    return null;
  }

  private Collection<Field> getAllDeclaredFields(Class<?> clazz) {
    List<Field> fields = new ArrayList<>();
    Collections.addAll(fields, clazz.getDeclaredFields());
    if (clazz.getSuperclass() != null) {
      fields.addAll(getAllDeclaredFields(clazz.getSuperclass()));
    }
    return fields;
  }
}
